// File: C:\cbproj\SoundUtl\SevenBitPackets.h
// Details, Purpose: See implementation, SevenBitPackets.c,
//   and C:\cbproj\SpecLab\html\Seven_Bit_Packets_for_the_serial_port.htm !
// Copyright, License, and Disclaimer : See SevenBitPackets.c .


#ifndef  _SEVEN_BIT_PACKETS_H_  // protect against multiple inclusion
# define _SEVEN_BIT_PACKETS_H_

// Compiling Spectrum Lab or a microcontroller firmware ?
#ifndef SBP_COMPILING_SPECLAB // if not defined elsewhere, make a guess:
# ifdef __BORLANDC__  // obviously compiling Spectrum Lab, because the compiler is Borland C
#  define SBP_COMPILING_SPECLAB 1 // 1 : "Compiling for Spectrum Lab", 0 : no
# else   // not compiling Spectrum Lab, guess we're compiling a microcontroller firmware..
#  define SBP_COMPILING_SPECLAB 0 // 0 because we're definitely NOT compiling Spectrum Lab.
# endif // compiling with Borland C ?
#endif // ndef SBP_COMPILING_SPECLAB ?

#if( SBP_COMPILING_SPECLAB ) // "Compiling Spectrum Lab" ?
#  define SBP_BUILT_IN_UNIT_TESTS 1 // 1 : implement the Unit Tests,        0 : don't
#  define SBP_USE_FLOATING_POINT  1 // 1 : may use 'float' and 'double',    0 : don't
#  define SBP_SUPPORT_PACKET_TX   1 // 1 : support TRANSMISSION of packets, 0 : don't
#  define SBP_SUPPORT_PACKET_RX   1 // 1 : support RECEPTION  of packets,   0 : don't
#  define SBP_MAX_AUDIO_CHANNELS  4 // 4 channels PER SAMPLE POINT, only used for internal buffering
#  define SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO 1 // 1 : use callbacks via FUNCTION POINTER, 0 : use SBP_AppCB_ProcessReceivedAudio()
#else   // not compiling Spectrum Lab, so guess we're compiling a microcontroller firmware,
   // which "only" drives an ADC (or read samples from its on-chip analog inputs),
   // but doesn't drive a DAC, thus it only needs to SEND but not RECEIVE packets
   // (all these are just guesses, so modify the following, or move these defs
   //  into a separate, "project specific" include file) .
#  define SBP_BUILT_IN_UNIT_TESTS 0 // 0 : don't implement the Unit Tests (save precious ROM..)
#  define SBP_USE_FLOATING_POINT  0 // 0 : neither use 'float' nor 'double' ("too expensive")
#  define SBP_SUPPORT_PACKET_TX   1 // 1 : support TRANSMISSION of packets, 0 : don't
#  define SBP_SUPPORT_PACKET_RX   0 // 0 : support RECEPTION  of packets,   0 : don't
#  define SBP_MAX_AUDIO_CHANNELS  2 // set this to the number of ADC- and/or DAC channels supported in your firmware.
#  define SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO 0 // 1 : use callbacks via FUNCTION POINTER, 0 : use SBP_AppCB_ProcessReceivedAudio()
#endif // SBP_COMPILING_SPECLAB ?


//----- Macro constants for the 'Seven Bit Packet' header bytes -------------


// Bitmasks and values for bitgroups in the packet_header_byte .
// Explained in Seven_Bit_Packets_for_the_serial_port.htm#packet_header_byte .
// Prefix "SB_PACKET_" means "applies to Seven Bit Packets" only :
#define SB_PACKET_HEADER_FLAG       0x80 // bit 7 : only set in the packet header byte
#define SB_PACKET_HDR_TYPE_MASK     0x60 // bits 6..5 : PACKET_HEADER_TYPE ...
#define SB_PACKET_HDR_TYPE_AUDIO    0x00 // b6..5 = 0b00 : 'short' audio sample point
#define SB_PACKET_HDR_TYPE_OTHER    0x20 // b6..5 = 0b01 : type specified by an extra byte,
        // that follows after the 14-bit "long packet length",
        // or (if it's a "short packet") immediately after the header byte.
#define SB_PACKET_HDR_TYPE_ASCII    0x40 // b6..5 = 0b10 : ASCII string, 7 bit per char
#define SB_PACKET_HDR_TYPE_RESERVED 0x60 // b6..5 = 0b11 : 'reserved for future use'
#define SB_PACKET_HDR_LENGTH_MASK   0x1F // b4..0 = FIVE-BIT "SHORT_PAYLOAD_LENGTH"
#define SB_PACKET_HDR_LENGTH_NOT_INDICATED 0 // special "SHORT_PAYLOAD_LENGTH"
        // for 'lazy packet transmission', e.g. for zero-terminated strings
        // if the microcontroller cannot afford to assemble the packet in memory
        // but write it 'directly into the UART transmit register'.
#define SB_PACKET_HDR_LENGTH_SHORT_DUMMY   1 // dummy for internal use - see
                                             // SBP_BeginWritingPacket() !
#define SB_PACKET_HDR_LENGTH_LONG_PAYLOAD 31 // "the length of this 'long' packet
        //  is indicated in two EXTRA bytes of the header, immediately after
        //  the header byte.

// 'Packet content types' for anything that is neither an AUDIO-packet
// nor a 7-bit ASCII string.
// Explained in Seven_Bit_Packets_for_the_serial_port.htm#packet_content_types :
#define SB_PACKET_CONTENT_TYPE_UNKNOWN              0x00
#define SB_PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT  0x01
#define SB_PACKET_CONTENT_TYPE_TIME_OF_DAY          0x02
#define SB_PACKET_CONTENT_TYPE_UNIX_DATE            0x03
#define SB_PACKET_CONTENT_TYPE_GPS_NMEA             0x04
// Negative dummy values for packet types indicated in two bits of the HEADER BYTE.
// Only used internally, for example as function argument for
#define SB_PACKET_CONTENT_TYPE_NO_EXTRA_BYTE        -1
#define SB_PACKET_CONTENT_TYPE_AUDIO                -2
#define SB_PACKET_CONTENT_TYPE_ASCII                -3
#define SB_PACKET_CONTENT_TYPE_OTHER                -4
#define SB_PACKET_CONTENT_TYPE_RESERVED             -5



//----- Data types (usually defined in WINDEF.H, stdint.h, etc) -------------

#ifndef  BYTE
# define BYTE unsigned char
#endif
#ifndef  DWORD
# define DWORD unsigned long
#endif
#ifndef  BOOL  // if not defined in WINDEF.H, define it here:
# define BOOL int
#endif
#ifndef  FALSE
# define FALSE 0
#endif
#ifndef  TRUE
# define TRUE  1
#endif

#ifndef  uint32_t // should have been defined in stdint.h or whatever ..
# define uint32_t unsigned long
#endif

#ifndef  uint16_t
# define uint16_t unsigned short
#endif

#ifndef  uint8_t
# define uint8_t unsigned char
#endif

#ifndef CPROT  // For peaceful co-existence of "C" and "C++" / call C-functions from C++ :
# ifdef __cplusplus
#   define CPROT extern "C"
# else
#   define CPROT
# endif  // ! "cplusplus"
#endif  // ndef CPROT


//----------------------------------------------------------------------------
// Callback function pointer for RECEIVED audio samples (invoked from the "Seven-Bit Packet Decoder")
typedef void (__stdcall *T_SBP_RxAudioCallback)( void *pvCallbackData, struct t_SBP_Instance *pInst );
         // ,------------------------------------------'                 |
         // '--> May point to whatever you like.                         |
         //      In Spectrum Lab, it's the address of a                  |
         //      'CSound' object that owns the T_SBP_Instance .          |
         // ,------------------------------------------------------------'
         // '--> Contains a single to-be-processed sample point,
         //      but also the number of channels PER SAMPLE POINT,
         //      and the precise sampling rate at which those sample-points
         //      'arrive here' on average.
         // As an alternative (without function pointers for the 'RX callback'),
         // #define SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO to ZERO,
         // and implement SBP_AppCB_ProcessReceivedAudio() in your application
         // (function prototype and details further below).


//----------------------------------------------------------------------------
typedef struct t_SBP_Instance // everything that belongs to a 'Seven-Bit Packetizer' instance
{  // (Spectrum Lab uses multiple of these: At least one for input, and one for output)
#if( SBP_COMPILING_SPECLAB )
#  define SBP_PACKET_BUFFER_SIZE 512 // long enough for a SHORT_PAYLOAD *and*
          // to parse RECEIVED or assemble TRANSMITTED "decently long" strings.
          // Reduce if necessary for a micro controller firmware with limited
          // resources (if strings are needed there at all).
#else  // not compiling for Spectrum Lab, but e.g. for a microcontroller with small RAM:
#  define SBP_PACKET_BUFFER_SIZE 32  // long enough for a SHORT_PAYLOAD only
#endif // SBP_COMPILING_SPECLAB ?
   BYTE bPacketBuffer[SBP_PACKET_BUFFER_SIZE]; // who needs a 'uint8_t' when we still have a BYTE ? :o)

   // Because any "Seven-Bit Payload" (SBP) may have to be filled in chunks (bitgroups)
   // of any size between 1 and 32 (bits of "payload"), a simple BYTE-pointer
   // isn't sufficient for writing bits to / reading bits from the above buffer.
   // Instead, the bit-fiddling functions SBP_ReadBitgroup() and SBP_WriteBitgroup()
   // use BIT-INDICES, not a BYTE-INDICES, to unpack or pack bitgroups from/to
   // blocks of bytes anywhere in RAM. The BYTE-POINTERS, pbSrc / pbDst, remain
   // unchanged as bitgroups are read from, or written to (appended) the payload.
   BYTE *pbSrc; // pointer to source with 7-Bit Payload  for SBP_ReadBitgroup()
   int iReadBitIndex;      // BIT- (not BYTE-) index used by SBP_ReadBitgroup()
   int nReadBitsAvailable; // "endstop" for iReadBitIndex in SBP_ReadBitgroup()

   BYTE *pbDst; // pointer to destination with 7-Bit Payload for SBP_WriteBitgroup()
   int iWriteBitIndex;    // BIT- (not BYTE-) index  used by SBP_WriteBitgroup()
   int nWriteBitsMax;     // "endstop" for iWriteBitIndex in SBP_WriteBitgroup()
   // Note: Having separate 'Source' and 'Destination' pointers here
   //       allows e.g. assembling a densely packed response in bPacketBuffer[],
   //       while analysing a received packet directly in the external caller's
   //       receive-buffer.


   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   // Variable format description of audio packets. At least Spectrum Lab can
   // not only SEND, but also RECEIVE (and parse) audio-packets, as well as
   // packets with PACKET_CONTENT_TYPE_AUDIO_SAMPLE_FORMAT . From the last
   // received packet of that type, the following struct members will be updated.
   // In the SENDER (e.g. microcontroller with an ADC and serial OUTPUT),
   // the struct members in this section must be filled out after the
   // application called SBP_InitInstance(). Don't waste ROM to pass all these
   // parameters 'nicely' as function arguments. Or even use a hard-coded
   // "audio format descriptor" (as a const BYTE array) in the firmware.
   int nBitsPerSingleSample; // typically 8, 14, 16, 21, or 24 bits per sample
       // in a single channel. The 'reference implementation' allows anything
       // from 1 to 32 bits.
   int nChannelsPerSamplePoint; // 1, 2, or maybe more channels per sample point
       // A packet with PACKET_HDR_TYPE_AUDIO contains ONE "sample point",
       // with the above number of channels, all sampled simultaneously.
   int iSampleDataType;  // sample data type and endianness. One of these:
#  define SBP_DATA_TYPE_SIGNED_LE    0  // signed integer, Little endian. 2 to 32 bits.
#  define SBP_DATA_TYPE_UNSIGNED_LE  1  // unsigned integer, Little endian. 1 to 32 bits.
#  define SBP_DATA_TYPE_SIGNED_BE    2  // reserved for signed integer, BIG endian.
#  define SBP_DATA_TYPE_UNSIGNED_BE  3  // reserved signed integer, BIG endian
#  define SBP_DATA_TYPE_FLOAT32_LE   4  // IEEE-754 32-bit floating point, Little Endian.
           // Value range normalized to -1.0 .. +1.0 for 'full swing from the A/C converter'.
           // The 'nBitsPerSingleSample' are ignored for this data type (always FOUR bytes per 'float').
#  define SBP_DATA_TYPE_FLOAT64_LE   8  // IEEE-754 64-bit floating point ("double"), LE,
           // Value range also normalized to -1.0 .. +1.0.  Useless for Spectrum Lab,
           // because the INTERNAL format used there (for signal processing)
           // is a single-precision "float".
# if(SBP_USE_FLOATING_POINT)
   double dblSampleRate_Hz; // Samples per second, as precise as possible.
   long double ldblUnixDateAndTime; // 'absolute' date and time in the next
          // sample (to be sent with the next SB_PACKET_CONTENT_TYPE_UNIX_DATE
          //         and/or SB_PACKET_CONTENT_TYPE_TIME_OF_DAY, both OPTIONAL).
          // Unit : "seconds passed since January 1st, 1970, 00:00:00.0" (in UTC).
          // Resolution: with Borland's "long double", which has 64 bit MANTISSA
          //     and a 15 bit exponent, the resolution is roughly 70 ps (!) :
          //     40 [years] * 365 * 24 * 60 * 60 / (2^64)  =  6.8e-11 [seconds]
# else  // for small microcontrollers without floating point support :
   long   i32SampleRate_Hz  // Samples per second, only integer values.
# endif // SBP_USE_FLOATING_POINT ?
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

#if(SBP_USE_FLOATING_POINT) // How to "buffer" a single sample point (with multiple channels) in RAM ?
   float fltSampleBuffer[ SBP_MAX_AUDIO_CHANNELS ]; // buffer to 'exchange' a single sample point with the application
         // Spectrum Lab uses floating point samples internally,
         // normalized to a value range of +/- 1.0, regardless of the ADC / DAC
         // resolution in "bits" (regardless of T_SBP_Instance.nBitsPerSingleSample).
         // That's what fltSampleBuffer[] uses, too.
#else  // ! SBP_USE_FLOATING_POINT : store a sample point as 'long' integers, not as normalized 'float' ->
   long i32SampleBuffer[ SBP_MAX_AUDIO_CHANNELS ]; // buffer to 'exchange' a single sample point with the application.
         // In this case ("without floats"), samples are neither scaled nor normalized.
         // They are only 'sign expanded' to 32 bits in the above buffer.
         // Make sure that T_SBP_Instance.nBitsPerSingleSample matches the
         // 'value range' in i32SampleBuffer[]. With the recommended data type
         // SBP_DATA_TYPE_SIGNED_LE, the value range in i32SampleBuffer should be..
         //      -32767 ...    +32767 for nBitsPerSingleSample = 16,
         //    -1048575 ...  +1048575 for nBitsPerSingleSample = 21,
         //   -16777215 ... +16777215 for nBitsPerSingleSample = 24, etc.
         //   |_______|---> Note the "missing codes" here. We don't want
         //                 the most NEGATIVE number to be e.g.   -32768
         //                 when the most POSITIVE number is e.g. +32767 .
         //   After expanding the variable-sized 'raw samples' to 32 bits,
         //   the only 'missing code', 0x80000000 = -2^31, is reserved to
         //   indicate 'invalid number' / 'no data' / 'error'.
#endif // SBP_USE_FLOATING_POINT ?


#if( SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO )
   T_SBP_RxAudioCallback pfRxAudioCallback; // 'function pointer', to be filled out by the application
   void *pvCallbackData;      // pointer to "anything" passed on to the above callback (in SL: a CSound object)
#endif // SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO ?

#if( SBP_COMPILING_SPECLAB )
   uint32_t dwRxPacketCounter, dwTxPacketCounter, dwErrorCounter; // only for 'statistics',
            // displayed in Spectrum Lab's "Debug"-window, "Misc"-tab or
            // "Comms"-tab (communication) after clicking 'Get a report'.

#endif // SBP_COMPILING_SPECLAB ?

} T_SBP_Instance;


//---------------------------------------------------------------------------
// Function prototypes (details only in the implementation, see *.c)
//---------------------------------------------------------------------------

CPROT void SBP_InitInstance( T_SBP_Instance *pInst );

// Bit-fiddling functions to read and write bitgroups in densely packed payload.
// Source or destination (7-Bit Payload) is in RAM (T_SBP_Instance.pbSrc, .pbDst).
CPROT uint32_t SBP_ReadBitgroup( T_SBP_Instance *pInst, int iBitgroupSize );
CPROT void     SBP_WriteBitgroup( T_SBP_Instance *pInst, uint32_t u32Value, int iBitgroupSize );

// To read bitgroups as SIGNED integers, here's another twin:
CPROT long SBP_ReadBitgroup_Signed( T_SBP_Instance *pInst, int iBitgroupSize );

// Conversion of floating point ("double", as internally used for the SAMPLING RATE,
//   for PACKET_CONTENT_TYPE_TIME_OF_DAY, etc into FIXED POINT with a 21(!) bit
//   signal fractional part (that can easily be 'left out' when not necessary):
#if( SBP_USE_FLOATING_POINT ) // May the implementation use 'float' and 'double' at all ?
CPROT void   SBP_DoubleToFixedPoint21( double d, long *pi32IntegerPart, long *pi32FractionalPart );
CPROT double SBP_FixedPoint21ToDouble( long integer_part, long fractional_part );
#endif // SBP_USE_FLOATING_POINT ?

// Packet assembly ("write into the internal packet buffer") :
CPROT int SBP_WritePacket_AudioSampleFormat(T_SBP_Instance *pInst);
CPROT int SBP_WritePacket_AudioSample(T_SBP_Instance *pInst);


#if( SBP_SUPPORT_PACKET_RX ) // how to pass RECEIVED AUDIO SAMPLES to the application ?
   // SBP_ProcessReceivedData(!) needs to 'call back' the application for
   //   EACH SAMPLE POINT contained in the block of received bytes.
   //   Especially under Windows, the number of bytes to process in a single
   //   call may be very high (due to the unpredictable timing and latencies),
   //   thus during a single call of SBP_ProcessReceivedData(), there may be
   //   dozens or even thousands of SAMPLE POINTS delivered to the 'application'
   //   (or microcontroller firmware driving a D/A-converter for output) !
   //  Depending on the target's capabilities, the delivery of multi-channel
   //  sample point to the application / uC firmware may use a callback
   //  via function pointer (type T_SBP_RxAudioCallback) or a normal function
   //  (SBP_AppCB_ProcessReceivedAudio(), located "somewhere in the APPLICATION").
   //  In any case, SBP_ProcessReceivedData() returns an INTEGER with the
   //       the number of bytes ACTUALLY CONSUMED (!) in the call,
   //       which is NOT NECESSARILY the same as the number of bytes
   //       'offered by the caller' (nBytesRcvd) !
   //       Details about why this is necessary, and what to do if
   //       SBP_ProcessReceivedData(!) cannot process ALL bytes ( return value
   //       less than nBytesRcvd) are in the implementation of
   //       that function, in SevenBitPackets.c .
# if( SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO )
   CPROT int SBP_ProcessReceivedData( // API function, called FROM the application
             // after the reception of ANY NUMBER OF BYTES from the serial port.
             T_SBP_Instance *pInst, BYTE *pbSource, int nBytesRcvd,
             T_SBP_RxAudioCallback pfRxAudioCallback, void *pvCallbackData );
# else // SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO == 0 ->
   // No callback via function pointer, but a normal 'subroutine call'.
   CPROT int SBP_ProcessReceivedData( // API function, called FROM the application
             // after the reception of ANY NUMBER OF BYTES from the serial port.
             T_SBP_Instance *pInst, BYTE *pbSource, int nBytesRcvd );
             // This is the 'simplistic' (and not very versatile) method
             // WITHOUT a function pointer to pass RECEIVED AUDIO SAMPLES
             // to the application. In this case, if the uC firmware supports
             // RECEPTION of packets at all, the following subroutine
             // must be implemented ***somewhere in the application*** :
   extern void SBP_AppCB_ProcessReceivedAudio( T_SBP_Instance *pInst );
        // ,----------------------------------------------------'
        // '--> Contains a single to-be-processed sample point,
        //      but also the number of channels PER SAMPLE POINT,
        //      and the precise sampling rate at which those sample-points
        //      'arrive here' on average.

# endif // ! SBP_USE_CALLBACK_FOR_RECEIVED_AUDIO ?
#endif // SBP_COMPILING_SPECLAB ?


#if( SBP_BUILT_IN_UNIT_TESTS )
CPROT int SBP_RunUnitTests( void ); // returns the NUMBER OF ERRORS, thus "ideally zero"
#endif // SBP_BUILT_IN_UNIT_TESTS ?


#endif // ndef _SEVEN_BIT_PACKETS_H_ ?

